#pragma once

#include <vector>
#include <utility>
#include <database/SqlValue.h>
#include <database/export.h>

namespace pdp {

/**
 * Sql result contains the result of the last database query.  The class
 * uses object types themselves that are implicitly shared (qt), so passing
 * around copies of this object should be fast.
 *
 * @ingroup database
 */
class PDPDATABASE_EXPORT SqlResult
{
public:

  /**
   * Structure for holding a row query result
   */
  struct QueryRow
  { std::vector<SqlValue> values; };

  SqlResult();

  virtual ~SqlResult();

  /**
   * This is true if an error was generated, or on an insert the row inserted
   * was found to be -1.
   */
  bool failure() const;

  /**
   * Increment the results to point to the next results row.  This follows the
   * access pattern of our old SqlQuery classes.
   */
  bool next() const
  {
    if (m_values.empty())
      return false;

    ++m_currentRow;
    if ((m_currentRow < (int)m_values.size()) && (m_currentRow >= 0))
      return true;
    else
    {
      m_currentRow = -1;
      return false;
    }
  }

  /**
   * Returns the value at index @a idx for the current result row.  This
   * function is used in conjunction with next().
   */
  const SqlValue& value(
    int idx) const
  {
    if (m_currentRow != -1)
      return m_values[m_currentRow].values[idx];
    else
      return m_null;
  }

  /**
   * Returns the value at row @a row and column @a col for the current result
   * set.
   */
  inline const SqlValue& value(
  	unsigned int row,
    unsigned int col) const
  { return m_values[row].values[col]; }

  /**
   * This method returns the number of rows obtained from the query.
   */
  unsigned int numRows() const;

  /**
   * Provides the number of columns retrieved in the query.
   */
  unsigned int numColumns() const;

  /**
   * Numbers of rows affected by the update or insert statement.  If no rows
   * were Affected, that means the updates where clause didn't match any rows in
   * the table.
   */
  unsigned int rowsAffected() const;

  /**
   * Last row id inserted by this SqlPreparedStatement.
   */
  unsigned int lastRowIdInserted() const;

  /**
   * Returns the error message from the last executed method.
   */
  const std::string errorMessage() const;

  /**
   * Used by the derived class to insert one row of results to the
   * class for access
   */
  void addRow(
    const QueryRow& row)
  { m_values.push_back(row); }

  /**
   * Used by the derived class to set any error messages received.
   */
  void setErrorMessage(
    const std::string& message);

  /**
   * Used by the derived class to identify the number of rows that were
   * affected by the sql query/operation.
   */
  void setNumRowsAffected(
    int numberOfRowsChanged);

  /**
   * Only set if the query was an insert.  A rowID of -1 indicates an error
   */
  void setLastRowIdInserted(int rowID);

private:
  std::vector<QueryRow> m_values;
  SqlValue m_null;
  mutable int m_currentRow;
  int m_numAffected;
  mutable std::string m_errorMessage;
  int m_lastInserted;
  bool m_isInsert;
};

} // namespace pdp

