#pragma once

#include <string>
#include <vector>
#include <boost/uuid/uuid.hpp>
#include <boost/uuid/string_generator.hpp>
#include <utilities/timeUtils.h>
#include <database/export.h>

namespace pdp {
  
/**
 * pdp Database SQL Value class
 *
 * @ingroup database
 */
class PDPDATABASE_EXPORT SqlValue
{
public:
  /**
   * SqlValue object constructor, creates a null Sql value.
   */
  explicit SqlValue()
    : m_isNull(true)
  { m_string.clear(); }

  /**
   * Creates an SqlValue that has binary data.  When created with this
   * constructor, only the toBinary will provide useful results.  All other
   * functions will cast an empty string to the type requested.
   */
  explicit SqlValue(
    const std::vector<unsigned char>& array)
    : m_isNull(false)
  { m_binary = array; }

  /**
   * Creates a SqlValue class with an QString.  Due to the implicit sharing
   * between QString's this constructor is preferrable over the const char*
   * constructor.  Basically it will mean that the same QString that the
   * database constructs is passed around without needless copies being made.
   */
  explicit SqlValue(
    const std::string& string)
    : m_isNull(false)
  { m_string = string; }

  /**
   * SqlValue object destructor.
   */
  ~SqlValue() {}

  /**
   * Clears the value
   */
  void clear()
  { m_string.clear(); }

  /**
   * Returns true if the value itself was a null in the database.
   */
  bool isNull() const
  { return m_isNull; }

  /**
   * Lets us know that the original data was binary data instead of a string.
   * Most of the converstion functions require the original data be of string
   * format.  Binary data can only be accessed with toBinary()
   */
  bool isBinary() const
  { return !m_binary.empty(); }

  /**
   * Return a copy of the current string
   */
  std::string string() const
  { return m_string; }

  /**
   * Converts the value to a boolean
   */
  bool toBool() const;

  /**
   * Converts the value to an int
   */
  int toInt() const
  { return std::stoi(m_string); }

  /**
   * Converts the value to a float
   */
  double toDouble() const
  { return std::stof(m_string); }
  
  /**
   * Returns the value as a boost::uuids::uuid
   */
  boost::uuids::uuid toUuid() const
  {
    boost::uuids::string_generator gen;
    return gen(m_string);
  }

 /**
  * Returns the value in the time format of 'yyyy.mm.dd-HH.MM.SS.fff' and returns a time object
  * set to the time value of the input string
  * 
  * @note This function requires that the value string have the form 'yyyy.mm.dd-HH.MM.SS.fff'.
  */
  std::chrono::system_clock::time_point toTimePoint() const
  { return stringToTimePoint(m_string); }

  /**
   * Will return an empty bytearray if the datatype is not stored as a 
   * binary in the database. Strings will not be converted to bytearrays.
   */
  const std::vector<unsigned char>& toByteArray() const
  { return m_binary; }

private:
  std::string m_string;
  std::vector<unsigned char> m_binary;
  bool m_isNull;
};
  
} // namespace pdp
