#pragma once

#include <string>
#include <sqlite3.h>
#include <database/SqlDatabase.h>
#include <database/export.h>

namespace pdp {

/**
 * Sqlite implementation of the pdp Database class
 *
 * @ingroup database
 */
class PDPDATABASE_EXPORT SqliteDatabase : public SqlDatabase
{
public:
  /**
   * SqliteDatabase object constructor.
   */
  SqliteDatabase();

  /**
   * SqliteDatabase object destructor.
   */
  virtual ~SqliteDatabase();

  // Virtual functions documented in base class
  virtual SqlDatabase::SqlDatabaseImplementation implementation() const override;
  
  virtual std::string blobTypeString() const override
  { return "BLOB"; }

  virtual std::string uuidTypeString() const override
  { return "VARCHAR(100)"; }

  virtual bool tableExists(
    const std::string& tableName) override;

  virtual bool tableColumnExists(
    const std::string& tableName,
    const std::string& columnName) override;

  virtual std::vector<std::string> columnNames(
    const std::string& tableName) override;

  /**
   * Unlike most databases where this command is supported.  In Sqlite it is
   * not, and the default isolation level is SERIALIZABLE
   */
  virtual bool supportsTransactionIsolationLevel() const override
  { return false; }

  virtual bool open(
    const std::string& connectionInfo,
    const std::string& password, 
    bool enableNotification,    ///< Ignored in an sqlite database
    bool applyLocalEncryption,
    bool inMemory
  ) override;

  /**
   * Open a SqliteDatabase. If the input database does not exist, it is
   * created and then opened. The connectionInfo is expected as a utf-8 encoded 
   * string. The encryption key is added when creating a new database. The
   * open will fail if opening an existing database with a different or no
   * encryption key.
   */
  bool openSqliteDatabase(
    const std::string& fileName,
    const std::vector<unsigned char>& password,
    bool encrypt,
    bool inMemory);

  virtual bool rekey(
    const std::string& fileName,
    const std::string& password) override;

  virtual bool isOpen() const override
  { return m_db ? true : false; }

  virtual void close() override;

  virtual const std::string errorMessage() const override
  { return m_errorMessage; }

  virtual SqlQuery* createQuery() override;

  virtual bool deleteDatabase(
    const std::string& databaseUsername,
    const std::string& databasePassword,
    const std::string& databaseName,
    const std::string& databasePath,
    const std::string& commandsIp,
    int commandsPort,
    std::string& errorMessage) override;

  virtual bool renameDatabase(
    const std::string& databaseUsername,
    const std::string& databasePassword,
    const std::string& databaseName,
    const std::string& databasePath,
    const std::string& commandsIp,
    int commandsPort,
    const std::string& newDatabaseName,
    std::string& errorMessage) override;

  virtual bool ensureConnected() override;

  sqlite3* db()
  { return m_db; }

  sqlite3_stmt* prepareStatement(
    const std::string& command);

protected:
  virtual SqlPreparedQuery* doCreatePreparedQuery(
    const std::string& sql) override;

private:
  std::string m_databaseFilePath;
  sqlite3* m_db;
  std::string m_errorMessage;
};

} // namespace pdp
