/** @file
 *
 * Regions composed of vectors with arbitrary element type.
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <elements/VectorR.h>
#include <elements/export.h>

namespace pdp {

/** 
 * A RegionR represents a multidimensional interval.  The
 * region is determined by the position of the corner with the lowest
 * coordinate in each dimension, and a nonnegative size.
 * 
 * @ingroup elements
 */
template<unsigned int N, typename E>
class RegionR
{
public:
  /**
   * Constructs an empty region. More precisely, the constructed region has
   * size zero and is positioned at the origin.
   */
  RegionR();

  /**
   * Constructs a RegionR from another region.
   */
  RegionR(
    const RegionR<N, E>& region);

  /**
   * Constructs a RegionR from a position and a size. 
   *
   * @pre all elements of @a size must be nonnegative
   */
  RegionR(
    const VectorR<N, E>& position,
    const VectorR<N, E>& size);

  /**
   * Assigns this RegionR from @a region.
   */
  RegionR<N, E>& operator=(
    const RegionR<N, E>& region);

  /**
   * Returns the position of this RegionR.
   */
   const VectorR<N, E>& position() const
  { return m_position; }

  /**
   * Sets the position and size of this RegionR.
   */
  inline void set(
    const VectorR<N, E>& position,
    const VectorR<N, E>& size)
  {
    m_position = position;
    m_size = size;
  }

  /**
   * Sets the position of this RegionR.
   */
  inline void setPosition(
    const VectorR<N, E>& position)
  { m_position = position; }

  /**
   * Returns the size of this RegionR.
   */
  inline const VectorR<N, E>& size() const
  { return m_size; }

  /**
   * Sets the size of this RegionR.
   * 
   * @pre all elements of @a size must be nonnegative
   */
  inline void setSize(
    const VectorR<N, E>& size)
  { m_size = size; }

  /**
   * Returns true if this RegionR is empty. For computed floating-point sizes,
   * this test is unreliable.
   */
  bool empty(
    double epsilon = PDP_EPSILON) const;

  /**
   * Sets the region to be empty.
   */
  void setEmpty();

  /**
   * Returns true if this RegionR equals @a region.
   */
  bool operator==(
    const RegionR& region) const;

  /**
   * Returns false if this RegionR equals @a region.
   */
  bool operator!=(
    const RegionR& region) const;

  /**
   * Returns true if this RegionR contains @a position. This treats the
   * region as open, meaning that the point must be less than all the
   * upper bounds to be considered inside
   */
  inline bool containsOpen(
    const VectorR<N, E>& position) const
  {
    for (unsigned int d = 0; d < N; ++d)
    {
      if (position[d] < m_position[d] || position[d] >= m_position[d] + m_size[d])
      {
        return false;
      }
    }
    return true;
  }

  /**
   * Returns true if this RegionR contains @a position. This treats the
   * region as closed, meaning that the point must be less than or equal
   * to all the upper bounds to be considered inside.
   */
  inline bool containsClosed(
    const VectorR<N, E>& position) const
  {
    for (unsigned int d = 0; d < N; ++d)
    {
      if (position[d] < m_position[d] || position[d] > m_position[d] + m_size[d])
      {
        return false;
      }
    }
    return true;
  }
  
  inline void increaseRegionBoundsToIncludePoint(
    const VectorR<N, E>& point)
  {
    for (unsigned int d = 0; d < N; ++d)
    {
      if (m_size[d] == (E)(0) && m_position[d] == (E)(0))
      {
        m_position[d] = point[d];
        m_size[d] = (E)(1);
      }
      if (point[d] < m_position[d])
      {
        m_size[d] = (m_position[d] + m_size[d]) - point[d];
        m_position[d] = point[d];
      }
      if (point[d] > m_position[d] + m_size[d])
      {
        m_size[d] = (point[d] - m_position[d]) + (E)(1);
      }
    }
  }

  //---------------------------------------------------------------------------
  // Friend functions
  //--------------------------------------------------------------------------- 
  /**
   * The method gets the next RegionR from the input stream and
   * decodes it from little endian format.
   */
  friend void getRegionR(
    std::istream& inputStream,
    RegionR<N, E>& r)
  {
    getVectorR(inputStream, r.m_position);
    getVectorR(inputStream, r.m_size);
  }

  /**
   * The method encodes the input RegionR into little endian format
   * and writes it to the output stream.
   */
  friend void putRegionR(
    const RegionR<N, E>& r,
    std::ostream& outputStream)
  {
    putVectorR(r.m_position, outputStream);
    putVectorR(r.m_size, outputStream);
  }
  
private:
  VectorR<N, E> m_position;
  VectorR<N, E> m_size;
};

/**
 * Prints @a region onto the output stream @a os.
 *
 * @relates RegionR
 */
template<unsigned int N, typename E>
std::ostream& operator<<(
  std::ostream& os,
  const RegionR<N, E>& region);

/**
 * Returns the intersection of two regions of the same dimesion and type.
 *
 * @relates RegionR
 */
template<unsigned int N, typename E>
const RegionR<N, E> intersection(
  const RegionR<N, E>& r1,
  const RegionR<N, E>& r2);

/**
 * Returns the ratio of the Intersection over the Union (IOU) of two regions of the same dimesions and type.
 *
 * @relates RegionR
 */
template<unsigned int N, typename E>
double intersectionOverUnion(
  const RegionR<N, E>& r1,
  const RegionR<N, E>& r2);

/**
 * Returns the bounding box of two regions of the same dimesion and type.
 *
 * @relates RegionR
 */
template<unsigned int N, typename E>
const RegionR<N, E> bounds(
  const RegionR<N, E>& r1,
  const RegionR<N, E>& r2);

/**
 * Returns the bounding box of a region and a vector of the same dimesion and type.
 *
 * @relates VectorR
 * @relates RegionR
 */
template<unsigned int N, typename E>
const RegionR<N, E> bounds(
  const RegionR<N, E>& r,
  const VectorR<N, E>& v);

/**
 * Returns an enlarged region of the specified region based on the margin size
 * (they both need to have the same dimension and type).
 *
 * @relates RegionR
 */
template<unsigned int N, typename E>
const RegionR<N, E> addMarginRegion(
  const RegionR<N, E>& primaryRegion,
  const VectorR<N, E>& marginSize);

/**
 * Constructs and returns a RegionR of dimension 2. The first two arguments
 * define the position of the region in the first and second dimension,
 * respectively, and the last two arguments define the corresponding size.
 *
 * @relates RegionR
 */
template<typename E>
inline RegionR<2, E> makeRegion(
  const VectorR<2, E>& position,
  const VectorR<2, E>& size)
{
  return RegionR<2, E>(position, size);
}

/**
 * Constructs and returns a RegionR of dimension 2. The first two arguments
 * define the position of the region in the first and second dimension,
 * respectively, and the last two arguments define the corresponding size.
 *
 * @relates RegionR
 */
template<typename E>
inline RegionR<2, E> makeRegion(
  const E& x,
  const E& y, 
  const E& xsize,
  const E& ysize)
{
  return RegionR<2, E>(makeVector(x, y), makeVector(xsize, ysize));
}

/**
 * Constructs and returns a RegionR of dimension 3. The first three
 * arguments define the position of the region in the first, second,
 * and third dimension, respectively, and the last three arguments
 * define the corresponding size.
 *
 * @relates RegionR
 */
template<typename E>
inline RegionR<3, E> makeRegion(
  const E& x,
  const E& y,
  const E& z, 
  const E& xsize,
  const E& ysize,
  const E& zsize)
{
  return RegionR<3, E>(makeVector(x, y, z), makeVector(xsize, ysize, zsize));
}

/**
 * Constructs and returns a RegionR of the specified type from the input
 * region by converting the type of each vector of the input region to
 * to the specified type. The regions must be composed of assigment compatible
 * vector types.
 *
 * @relates RegionR
 */
template<unsigned int N, typename E1, typename E2>
inline RegionR<N, E2> convertToRegionR(
  const RegionR<N, E1>& r)
{
  return RegionR<N, E2>(convertToVectorR<N, E1, E2>(r.position()),
    convertToVectorR<N, E1, E2>(r.size()));
}

/**
 * Specialization of the @a convertToVectorR function for converting
 * float to integer regions
 *
 * @relates RegionR
 */
template<unsigned int N>
inline RegionR<N, int> convertToRegionRi(
  const RegionR<N, float> r)
{
  return RegionR<N, int>(convertToVectorRi<N>(r.position()),
    convertToVectorRi<N>(r.size()));
}

/**
 * Specialization of the @a convertToVectorR function for converting
 * double to integer regions
 *
 * @relates RegionR
 */
template<unsigned int N>
inline RegionR<N, int> convertToRegionRi(
  const RegionR<N, double> r)
{
  return RegionR<N, int>(convertToVectorRi<N>(r.position()),
    convertToVectorRi<N>(r.size()));
}

/*
 * Short type names for common RegionR specializations.
 */
typedef RegionR<2, int> RegionR2i;
typedef RegionR<2, float> RegionR2f;
typedef RegionR<2, double> RegionR2d;
typedef RegionR<3, int> RegionR3i;
typedef RegionR<3, float> RegionR3f;
typedef RegionR<3, double> RegionR3d;

} // namespace pdp

#include <elements/RegionR.t>

