namespace pdp {

template<unsigned int N, typename E>
RegionR<N, E>::RegionR()
{}

template<unsigned int N, typename E>
RegionR<N, E>::RegionR(
  const RegionR<N, E>& region)
  : m_position(region.m_position),
    m_size(region.m_size)
{}

template<unsigned int N, typename E>
RegionR<N, E>::RegionR(
  const VectorR<N, E>& position,
  const VectorR<N, E>& size)
  : m_position(position),
    m_size(size)
{
  assert(length(m_size) >= 0);
}

template<unsigned int N, typename E>
RegionR<N, E>& RegionR<N, E>::operator=(
  const RegionR<N, E>& region)
{
  m_position = region.position();
  m_size = region.size(); // always non-negative
  return *this;
}

template<unsigned int N, typename E>
bool RegionR<N, E>::empty(
  double epsilon) const
{
  return (almostEqual(length(m_size), 0.0, epsilon));
}

template<unsigned int N, typename E>
void RegionR<N, E>::setEmpty()
{
  m_position = VectorR<N, E>(static_cast<E>(0.0));
  m_size = VectorR<N, E>(static_cast<E>(0.0));
}

template<unsigned int N, typename E>
bool RegionR<N, E>::operator==(
  const RegionR& region) const
{
  return (position() == region.position() && size() == region.size());
}

template<unsigned int N, typename E>
bool RegionR<N, E>::operator!=(
  const RegionR& region) const
{
  return !(*this == region);
}

template<unsigned int N, typename E>
std::ostream& operator<<(
  std::ostream& os,
  const RegionR<N, E>& region)
{
  os << region.position() << ' ' << region.size();
  return os;
}

template<unsigned int N, typename E>
const RegionR<N, E> intersection(
  const RegionR<N, E>& r1,
  const RegionR<N, E>& r2)
{
  VectorR<N, E> lower, size;
  for (unsigned int i = 0; i < N; ++i)
  {
    lower[i] = std::max(r1.position()[i], r2.position()[i]);
    E upper = std::min(r1.position()[i] + r1.size()[i], r2.position()[i] + r2.size()[i]);
    size[i] = upper - lower[i];
    if (size[i] < 0)
      size[i] = 0;
  }
  return RegionR<N, E>(lower, size);
}

template<unsigned int N, typename E>
double intersectionOverUnion(
  const RegionR<N, E>& r1,
  const RegionR<N, E>& r2)
{
  RegionR<N, E> inner = intersection(r1, r2);
  double fillOfR1 = 1.0f;
  double fillOfR2 = 1.0f;
  double fillOfInner = 1.0f;
  for (unsigned int i = 0; i < N; ++i)
  {
    fillOfR1 *= (double)r1.size()[i];
    fillOfR2 *= (double)r2.size()[i];
    fillOfInner *= (double)inner.size()[i];
  }
#ifdef _WINDOWS
#pragma warning( push )
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif
  return almostEqual((fillOfR1 + fillOfR2 - fillOfInner), 0.0, 0.0001) ?
    0.0 : fillOfInner / (fillOfR1 + fillOfR2 - fillOfInner);
#ifdef _WINDOWS
#pragma warning( pop )
#endif
}

template<unsigned int N, typename E>
const RegionR<N, E> bounds(
  const RegionR<N, E>& r1,
  const RegionR<N, E>& r2)
{
  VectorR<N, E> lower, size;
  for (unsigned int i = 0; i < N; ++i)
  {
    lower[i] = std::min(r1.position()[i], r2.position()[i]);
    E upper = std::max(r1.position()[i] + r1.size()[i], r2.position()[i] + r2.size()[i]);
    size[i] = upper - lower[i];
  }
  return RegionR<N, E>(lower, size);
}

template<unsigned int N, typename E>
const RegionR<N, E> bounds(
  const RegionR<N, E>& r,
  const VectorR<N, E>& v)
{
  VectorR<N, E> lower, size;
  for (unsigned int i = 0; i < N; ++i)
  {
    lower[i] = std::min(r.position()[i], v[i]);
    E upper = std::max(r.position()[i] + r.size()[i], v[i]);
    size[i] = upper - lower[i];
  }
  return RegionR<N, E>(lower, size);
}

template<unsigned int N, typename E>
const RegionR<N, E> addMarginRegion(
  const RegionR<N, E>& primaryRegion,
  const VectorR<N, E>& marginSize)
{
  E* pos = new E[N];
  E* size = new E[N];

  for (unsigned int i = 0; i < N; ++i)
  {
    pos[i] = primaryRegion.position()[i]-(marginSize[i]/((E)2));
    size[i] = primaryRegion.size()[i] + marginSize[i];
  }

  RegionR<N, E> tmpRegion(pos, size);

  delete [] pos;
  delete [] size;

  return tmpRegion;
}

} // namespace pdp
