namespace pdp {

template<unsigned int N, typename E>
void TriangleListR<N, E>::addVertex(
  bool addProcessingElements,
  const VectorR<N, E>& p,
  const VectorR<N, E>& n)
{
  // Add the vertex to the vertex list
  m_positions.push_back(p);
  m_normals.push_back(n);

  // Add the processing elements if called for
  if (addProcessingElements)
  {
    // Add empty triangle and neighbor vertex lists for the new vertex
    IndexList indexList;
    m_triangleLists.push_back(indexList);
    m_neighborVertexLists.push_back(indexList);
  }
}

template<unsigned int N, typename E>
void TriangleListR<N, E>::vertex(
  unsigned int idx,
  VectorR<N, E>& p,
  VectorR<N, E>& n) const
{
  if(idx < m_positions.size())
  {
    p.setElements(m_positions[idx].elements());
    n.setElements(m_normals[idx].elements());
  }
  else
  {
    p.set((E)0);
    n.set((E)0);
  }
}

template<unsigned int N, typename E>
void TriangleListR<N, E>::addTriangle(
  bool addProcessingElements,
  unsigned int v0Idx,
  unsigned int v1Idx,
  unsigned int v2Idx)
{
  // Check the bounds
  if (v0Idx >= m_positions.size() || v1Idx >= m_positions.size() || v2Idx >= m_positions.size())
    return;
  
  // Add the input vertex indexes to the triangle
  Triangle triangle;
  triangle.v0 = v0Idx;
  triangle.v1 = v1Idx;
  triangle.v2 = v2Idx;

  // Add the processing elements if called for
  if (addProcessingElements)
  {
    // Add the triangle index to the triangle lists for each vertex
    m_triangleLists[v0Idx].indexes.push_back(static_cast<unsigned int>(m_triangles.size()));
    m_triangleLists[v1Idx].indexes.push_back(static_cast<unsigned int>(m_triangles.size()));
    m_triangleLists[v2Idx].indexes.push_back(static_cast<unsigned int>(m_triangles.size()));

    // Add the neighbor vertex indexes to the neighbor vertex list for vertex 0
    bool inList = false;
    for (unsigned int i = 0; i < m_neighborVertexLists[v0Idx].indexes.size(); i++)
    {
      if (v1Idx == m_neighborVertexLists[v0Idx].indexes[i])
      {
        inList = true;
        break;
      }
    }
    if (!inList)
      m_neighborVertexLists[v0Idx].indexes.push_back(v1Idx);
    inList = false;
    for (unsigned int i = 0; i < m_neighborVertexLists[v0Idx].indexes.size(); i++)
    {
      if (v2Idx == m_neighborVertexLists[v0Idx].indexes[i])
      {
        inList = true;
        break;
      }
    }
    if (!inList)
      m_neighborVertexLists[v0Idx].indexes.push_back(v2Idx);

    // Add the neighbor vertex indexes to the neighbor vertex list for vertex 1
    inList = false;
    for (unsigned int i = 0; i < m_neighborVertexLists[v1Idx].indexes.size(); i++)
    {
      if (v0Idx == m_neighborVertexLists[v1Idx].indexes[i])
      {
        inList = true;
        break;
      }
    }
    if (!inList)
      m_neighborVertexLists[v1Idx].indexes.push_back(v0Idx);
    inList = false;
    for (unsigned int i = 0; i < m_neighborVertexLists[v1Idx].indexes.size(); i++)
    {
      if (v2Idx == m_neighborVertexLists[v1Idx].indexes[i])
      {
        inList = true;
        break;
      }
    }
    if (!inList)
      m_neighborVertexLists[v1Idx].indexes.push_back(v2Idx);

    // Add the neighbor vertex indexes to the neighbor vertex list for vertex 2
    inList = false;
    for (unsigned int i = 0; i < m_neighborVertexLists[v2Idx].indexes.size(); i++)
    {
      if (v0Idx == m_neighborVertexLists[v2Idx].indexes[i])
      {
        inList = true;
        break;
      }
    }
    if (!inList)
      m_neighborVertexLists[v2Idx].indexes.push_back(v0Idx);
    inList = false;
    for (unsigned int i = 0; i < m_neighborVertexLists[v2Idx].indexes.size(); i++)
    {
      if (v1Idx == m_neighborVertexLists[v2Idx].indexes[i])
      {
        inList = true;
        break;
      }
    }
    if (!inList)
      m_neighborVertexLists[v2Idx].indexes.push_back(v1Idx);
    m_triangleNormals.push_back(m_normals[v0Idx]);
  }

  // Add the new triangle to the triangle table
  m_triangles.push_back(triangle);
}

template<unsigned int N, typename E>
void TriangleListR<N, E>::triangle(
  unsigned int idx,
  VectorR<N, E>& p0,
  VectorR<N, E>& n0,
  VectorR<N, E>& p1,
  VectorR<N, E>& n1,
  VectorR<N, E>& p2,
  VectorR<N, E>& n2) const
{
  if(idx < m_triangles.size())
  {
    p0.setElements(m_positions[m_triangles[idx].v0].elements());
    n0.setElements(m_normals[m_triangles[idx].v0].elements());
    p1.setElements(m_positions[m_triangles[idx].v1].elements());
    n1.setElements(m_normals[m_triangles[idx].v1].elements());
    p2.setElements(m_positions[m_triangles[idx].v2].elements());
    n2.setElements(m_normals[m_triangles[idx].v2].elements());
  }
  else
  {
    p0.set((E)0);
    n0.set((E)0);
    p1.set((E)0);
    n1.set((E)0);
    p2.set((E)0);
    n2.set((E)0);
  }
}

template<unsigned int N, typename E>
void TriangleListR<N, E>::lowpass(
  double lamda,
  double sigma,
  unsigned int numberOfIterations)
{
  // Check for valid neighbor vertex lists
  if (m_neighborVertexLists.size() == 0)
    return;

  // Container for the adjusted vertexes
  std::vector< VectorR<N, E> > newPositions;
  std::vector< VectorR<N, E> > newNormals;
  VectorR<N, E> newPosition;
  VectorR<N, E> newNormal;
  E weight;
  VectorR<N, E> delta, deltaSum;
  for (unsigned int pass = 0; pass < numberOfIterations; pass++)
  {
    // Loop over the vertexes of the triangle surface and apply the positive smoothing correction
    for (unsigned int i = 0; i < m_positions.size(); i++)
    {
      // Compute the delta from the weighted average of the current vertex neighbors
      deltaSum.set((E)(0.0));
      if (m_neighborVertexLists[i].indexes.size() > 0)
      {
        weight = (E)(lamda / (double)m_neighborVertexLists[i].indexes.size());
        for (unsigned int j = 0; j < m_neighborVertexLists[i].indexes.size(); j++)
        {
          sub(m_positions[m_neighborVertexLists[i].indexes[j]], m_positions[i], delta);
          add(deltaSum, delta, deltaSum);
        }
        mult(deltaSum, weight, deltaSum);
      }

      // Compute the new vertex and add to the list
      newNormal = m_normals[i];
      add(m_positions[i], deltaSum, newPosition);
      newPositions.push_back(newPosition);
      newNormals.push_back(newNormal);
    }

    // Clear the current vertexes list and add the new points
    m_positions.clear();
    m_normals.clear();
    for (unsigned int i = 0; i < newPositions.size(); i++)
    {
      m_positions.push_back(newPositions[i]);
      m_normals.push_back(newNormals[i]);
    }
    newPositions.clear();
    newNormals.clear();

    // Loop over the vertexes of the triangle surface and apply the negative smoothing correction
    for (unsigned int i = 0; i < m_positions.size(); i++)
    {
      // Compute the delta from the weighted average of the current vertex neighbors
      deltaSum.set((E)(0.0));
      if (m_neighborVertexLists[i].indexes.size() > 0)
      {
        weight = (E)(sigma / (double)m_neighborVertexLists[i].indexes.size());
        for (unsigned int j = 0; j < m_neighborVertexLists[i].indexes.size(); j++)
        {
          sub(m_positions[m_neighborVertexLists[i].indexes[j]], m_positions[i], delta);
          add(deltaSum, delta, deltaSum);
        }
        mult(deltaSum, weight, deltaSum);
      }

      // Compute the new vertex and add to the list
      newNormal = m_normals[i];
      add(m_positions[i], deltaSum, newPosition);
      newPositions.push_back(newPosition);
      newNormals.push_back(newNormal);
    }

    // Clear the current vertexes list and add the new points
    m_positions.clear();
    m_normals.clear();
    for (unsigned int i = 0; i < newPositions.size(); i++)
    {
      m_positions.push_back(newPositions[i]);
      m_normals.push_back(newNormals[i]);
    }
    newPositions.clear();
    newNormals.clear();
  }
}

template<unsigned int N, typename E>
void TriangleListR<N, E>::computeNormals()
{
  // Check for valid triangles lists
  if (m_triangleLists.size() == 0 || m_triangleNormals.size() == 0)
    return;

  // Loop over the triangles and compute the normals for each triangle
  TriangleR<N, E> triangle;
  VectorR<N, E> zeroNormal;
  zeroNormal.set((E)(0.0));
  for (unsigned int i = 0; i < m_triangles.size(); i++)
  {
    triangle.set(m_positions[m_triangles[i].v0], m_positions[m_triangles[i].v1],
      m_positions[m_triangles[i].v2], zeroNormal, zeroNormal, zeroNormal);
    triangle.computeNormal();
    m_triangleNormals[i].setElements(triangle.n0().elements());
  }

  // Loop over the vertexes and compute their normals from the average of the triangle normals
  VectorR<N, E> normalSum;
  for (unsigned int i = 0; i < m_positions.size(); i++)
  {
    // Loop over the triangles for the current vertex and compute the their normal average
    normalSum.set((E)(0.0));
    for (unsigned int j = 0; j < m_triangleLists[i].indexes.size(); j++)
      add(m_triangleNormals[m_triangleLists[i].indexes[j]], normalSum, normalSum);
    normalize(normalSum);
    m_normals[i] = normalSum;
  }
}

} // namespace pdp
