/** @file
 *
 * Real Vector utility classes and functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <ostream>
#include <string>
#include <vector>
#include <math.h>
#include <algorithm>
#include <utilities/stringUtils.h>
#include <elements/pdpmath.h>

namespace pdp {

/**
 * The VectorR template class represents fixed-size, real vectors.  The
 * size of the vector and the element type are specified as part of the
 * type, and are thus fixed at compile time. This vector class represents
 * both positional vectors representing points, and covariance vectors such
 * as normal and tangent vectors.
 *
 * @a N is the number of elements, which must be greater than zero.
 * @a E is the element type, typically @c double, @c float, or @c int.
 *
 * @ingroup elements
 */
template<unsigned int N, typename E>
class VectorR
{
public:
  /**
   * The number of elements in vectors of this type.
   */
  static const unsigned int SIZE = N;

  typedef E value_type;

  /**
   * Constructs a vector whose elements are uninitialized.
   */
  inline VectorR()
  { std::fill(&(m_element[0]), &(m_element[0]) + N, static_cast<E>(0)); }

  /**
   * Destroys a vector.
   */
  inline ~VectorR() = default;

  /**
   * Constructs a vector and sets all elements to @a element.
   */
  VectorR(
    E element);

  /**
   * Added c++11 initializer list support.
   * @example VectorR3f one = {1.0f, 2.0f, 3.0f};
   */
  VectorR(
    std::initializer_list<E> l)
  {
    for (size_t i = 0; i < N; ++i)
      m_element[i] = *(l.begin() + i);
  }

  /**
   * Constructs a vector and copies the elements from @a elementArray.
   * The length of the input array must be at least as long as the vector.
   */
  VectorR(
    const E* elementArray);

  /**
   * Constructs a copy of @a v.
   */
  VectorR(
    const VectorR<N, E>& v) = default;

  /**
   * Returns a pointer to an array containing the elements of this vector.
   * The caller must not try to deallocate this array.
   */
  const E* elements() const
  {
    return m_element;
  }

  /**
   * Returns a pointer to an array containing the elements of this vector.
   * The caller must not try to deallocate this array. Assigning elements
   * of this array will change the vectorR
   */
  E* elements()
  {
    return m_element;
  }

  /**
   * Sets all elements of this vector to @a element.
   */
  void set(
    const E& element)
  {
    std::fill(m_element, m_element + N, element);
  }

  /**
   * Copies the elements from @a elementArray to this vector.  The
   * array must contain @a N elements.
   */
  void setElements(
    const E* elementArray)
  {
    std::copy(elementArray, elementArray + N, m_element);
  }

  /**
   * Assigns this Vector from another vector @a v.
   */
  VectorR<N, E>& operator=(const VectorR<N, E>& v)
  {
    setElements(v.elements());
    return *this;
  }

  /**
   * Returns a writable reference to element @a idx.
   *
   * @pre 0 <= idx < N
   */
  E& operator[](
    unsigned int idx)
  {
    return m_element[idx];
  }

  /**
   * Returns a constant reference to element @a idx.
   *
   * @pre 0 <= idx < N
   */
  const E operator[](
    unsigned int idx) const
  {
    return m_element[idx];
  }

  //---------------------------------------------------------------------------
  // Friend functions
  //---------------------------------------------------------------------------
  /**
   * The method gets the next vector from the input stream and
   * decodes it from little endian format.
   */
  friend void getVectorR(
    std::istream& inputStream,
    VectorR<N, E>& v)
  { for (unsigned int i = 0; i < N; i++) v.m_element[i] = getR<E>(inputStream); }

  /**
   * The method encodes the input vector into little endian format
   * and writes it to the output stream.
   */
  friend void putVectorR(
    const VectorR<N, E>& v,
    std::ostream& outputStream)
  { for (unsigned int i = 0; i < N; i++) putR<E>(v.m_element[i], outputStream); }

private:
  // member variables - note opengl code assumes no other state is in a VectorR
  E m_element[N];
};

//---------------------------------------------------------------------------
// Related functions
//---------------------------------------------------------------------------
/**
 * Reports the value of the vector @a v in text format to
 * the output stream @a out
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
std::ostream& operator<<(
  std::ostream& out,
  const VectorR<N, E> & v)
{
  out << "<";
  for (unsigned int k = 0; k < N-1; ++k)
  {
    out << v[k] << ", ";
  }
  out << v[N-1] << ">";
  return out;
}

/**
 * Returns the sum between @a v1 and @a v2.  Note, there is a significant
 * performance penalty to using this operator in vs2003 vs vs2005.  This is due
 * the fact that the Named Return Value optimization was not implemented until
 * Visual Studio 2005.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
const VectorR<N, E> operator+(
  const VectorR<N, E>& v1,
  const VectorR<N, E>& v2)
{
  VectorR<N, E> result;
  add(v1, v2, result);
  return result;
}

/**
 * Returns the difference between @a v1 and @a v2. Note, there is a significant
 * performance penalty to using this operator in vs2003 vs vs2005.  This is due
 * the fact that the Named Return Value optimization was not implemented until
 * Visual Studio 2005.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
const VectorR<N, E> operator-(
  const VectorR<N, E>& v1,
  const VectorR<N, E>& v2)
{
  VectorR<N, E> result;
  sub(v1, v2, result);
  return result;
}

/**
 * Returns the negative of the input vector @a
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
const VectorR<N, E> operator-(
  const VectorR<N, E>& v)
{
  VectorR<N, E> result;
  mult(v, static_cast<E>(-1.0), result);
  return result;
}

/**
 * Returns a vector which is the input vector @a v scaled by the
 * input scale @a scaleValue
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
const VectorR<N, E> operator*(
  const VectorR<N, E>& v,
  const E multValue)
{
  VectorR<N, E> result;
  mult(v, multValue, result);
  return result;
}

/**
 * Constructs and returns a Vector of length 2, where the first
 * element is set to @a x and the second to @a y.
 *
 * @relates VectorR
 */
template<typename E>
inline VectorR<2, E> makeVector(
  const E& x,
  const E& y)
{
  VectorR<2, E> v;
  v[0] = x;
  v[1] = y;
  return v;
}

/**
 * Sets to a unit vector if the magnitude of the input vector is
 * greater than @a epsilon
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
void normalize(
  VectorR<N, E>& v,
  double epsilon)
{
  auto len = length(v);
  if (almostZero((E)length(v), (E)epsilon))
  {
    for (unsigned int k = 0; k < N; ++k)
    {
      v[k] = static_cast<E>(0.0);
    }
  }
  else
  {
    mult(v, static_cast<E>(1.0 / len), v);
  }
}

/**
 * Adds input vector @a v0 to input vector @a v1 and stores the
 * result in the result vector @a result. For this function it is ok if
 * the result vector and either of the input vectors refer to the
 * same object. This function can only be generated for input and result
 * vectors with the same dimension.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline void add(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1,
  VectorR<N, E>& result)
{
  for (unsigned int i = 0; i < N; ++i)
    result[i] = v0[i] + v1[i];
}

/**
 * Subtracts input vector @a v1 from input vector @a v0 and stores the
 * result in the result vector @a result. For this function it is ok if
 * the result vector and either of the input vectors refer to the
 * same object. This function can only be generated for input and result
 * vectors with the same dimension.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline void sub(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1,
  VectorR<N, E>& result)
{
  for (unsigned int i = 0; i < N; ++i)
    result[i] = v0[i] - v1[i];
}

/**
 * Scales input vector @a v by input value @a b and stores the
 * result in the result vector @a result. For this function it is ok if
 * the result vector and the input vector refer to the same object. This
 * function can only be generated for input and result vectors with the same dimension.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline void mult(
  const VectorR<N, E>& v,
  const E scale,
  VectorR<N, E>& result)
{
  for (unsigned int i = 0; i < N; ++i)
    result[i] = v[i] * scale;
}

/**
 * Scales input vector @a v0 by input vector @a v1 and stores the
 * result in the result vector @a result. For this function it is ok if
 * the result vector and either of the input vectors refer to the
 * same object. This function can only be generated for input and result
 * vectors with the same dimension.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline void mult(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1,
  VectorR<N, E>& result)
{
  for (unsigned int i = 0; i < N; ++i)
    result[i] = v0[i] * v1[i];
}

/**
 * Divides each element in input vector @a v0 by input vector @a v1 and stores the
 * result in the result vector @a result. For this function it is ok if
 * the result vector and either of the input vectors refer to the
 * same object. This function can only be generated for input and result
 * vectors with the same dimension.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline void divide(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1,
  VectorR<N, E>& result)
{
  for (unsigned int i = 0; i < N; ++i)
    result[i] = v0[i] / v1[i];
}

/**
 * Creates the average of in input vector @a v0 and input vector @a v1 and stores the
 * result in the result vector @a result. For this function it is ok if
 * the result vector and either of the input vectors refer to the
 * same object. This function can only be generated for input and result
 * vectors with the same dimension.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline void average(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1,
  VectorR<N, E>& result)
{
  for (unsigned int i = 0; i < N; ++i)
    result[i] = (v0[i] + v1[i]) / (E)2;
}

/**
 * Returns @a true if all the individual elements of the input vector
 * @a v are within @a epsilon of zero
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
bool almostZero(
  const VectorR<N, E>& v,
  double epsilon)
{
  for (unsigned int k = 0; k < N; ++k)
  {
    if (!almostZero(v[k], epsilon))
      return false;
  }
  return true;
}

/**
 * Returns @a true if the input vector @a v is within @a epsilon of being
 * a unit vector
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
bool almostUnit(
  const VectorR<N, E>& v,
  double epsilon)
{
  return(almostZero(static_cast<E>(1.0) - length(v), epsilon));
}

/**
 * Returns @a true if the per-element difference between input vector @a v0
 * and input vector @a v1 is within @a epsilon.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
bool almostEqual(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1,
  double epsilon)
{
  for (unsigned int k = 0; k < N; ++k)
  {
#ifdef _WINDOWS
#pragma warning( push )
#pragma warning( disable : 4244 )
#endif
    if (!almostZero(v0[k] - v1[k], epsilon))
      return false;
#ifdef _WINDOWS
#pragma warning( pop )
#endif
  }
  return true;
}

/**
 * Returns the dot product of the input vectors @a v0 and @a v1
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
double dot(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1)
{
  double result = 0;
  for (unsigned int k = 0; k < N; ++k)
  {
    result += v0[k] * v1[k];
  }
  return result;
}

/**
 * Returns the angle, in radians, between the input vectors @a v0 and @a v1
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
double angle(
  const VectorR<N, E>& v0,
  const VectorR<N, E>& v1)
{
  double dotProduct = 0;
  for (unsigned int k = 0; k < N; ++k)
  {
    dotProduct += v0[k] * v1[k];
  }
  double arg = dotProduct / (length(v0) * length(v1));
  if(arg > 1.0) arg = 1.0;
  if(arg < -1.0) arg = -1.0;
  return acos(arg);
}

/**
 * Calculates the vector cross product of input vector @a v0 with input
 * vector @a v1 and stores the result in the result vector @a result.
 *
 * @relates VectorR
 */
template<typename E>
void cross(
  const VectorR<3, E>& v0,
  const VectorR<3, E>& v1,
  VectorR<3, E>& result)
{
  result[0] = v0[1]*v1[2] - v0[2]*v1[1];
  result[1] = v0[2]*v1[0] - v0[0]*v1[2];
  result[2] = v0[0]*v1[1] - v0[1]*v1[0];
}

/**
 * Returns the length of the vector.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
double length(
  const VectorR<N, E>& v)
{ return sqrt(dot(v, v)); }

/**
 * Constructs and returns a VectorR of the specified type from the input
 * vector by converting the type of each element of the input to the
 * specified type. The vectors must be composed of assigment compatible
 * element types.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E1, typename E2>
inline VectorR<N, E2> convertToVectorR(
  const VectorR<N, E1>& v)
{
  VectorR<N, E2> result;
  for (unsigned int i = 0; i < N; i++)
  {
    result[i] = static_cast<E2>(v[i]);
  }
  return result;
}

/**
 * Specialization of the @a convertToVectorR function for converting
 * float to integer vectors
 *
 * @relates VectorR
 */
template<unsigned int N>
inline VectorR<N, int> convertToVectorRi(
  const VectorR<N, float>& v)
{
  VectorR<N, int> result;
  for (unsigned int i = 0; i < N; i++)
  {
    result[i] = nearestInt(v[i]);
  }
  return result;
}

/**
 * Specialization of the @a convertToVectorR function for converting
 * double to integer vectors
 *
 * @relates VectorR
 */
template<unsigned int N>
inline VectorR<N, int> convertToVectorRi(
  const VectorR<N, double>& v)
{
  VectorR<N, int> result;
  for (unsigned int i = 0; i < N; i++)
  {
    result[i] = nearestInt(v[i]);
  }
  return result;
}

/**
 * Constructs and returns a Vector of length 3, where the first
 * element is set to @a x, the second to @a y, and the third to @a z.
 * The element type is deduced from the argument type.
 *
 * @relates VectorR
 */
template<typename E>
inline VectorR<3, E> makeVector(
  const E& x,
  const E& y,
  const E& z)
{
  VectorR<3, E> v;
  v[0] = x;
  v[1] = y;
  v[2] = z;
  return v;
}

/**
 * Constructs and returns a Vector of length 4, where the first
 * element is set to @a x, the second to @a y, the third to @a z, and
 * the forth to @a w. The element type is deduced from the argument
 * type.
 *
 * @relates VectorR
 */
template<typename E>
inline VectorR<4, E> makeVector(
  const E& x,
  const E& y,
  const E& z,
  const E& w)
{
  VectorR<4, E> v;
  v[0] = x;
  v[1] = y;
  v[2] = z;
  v[3] = w;
  return v;
}

/**
 * Constructs and returns a Vector of length N with the specified default value.  The element type and length
 * are deduced from the argument types.
 *
 * @relates VectorR
 */
template<unsigned int N, typename E>
inline VectorR<N, E> makeVector(
  const E& defaultValue)
{
  VectorR<N, E> v(defaultValue);
  return v;
}

typedef VectorR<2, int> VectorR2i;
typedef VectorR<3, int> VectorR3i;
typedef VectorR<4, int> VectorR4i;
typedef VectorR<2, unsigned int> VectorR2ui;
typedef VectorR<3, unsigned int> VectorR3ui;
typedef VectorR<4, unsigned int> VectorR4ui;
typedef VectorR<2, float> VectorR2f;
typedef VectorR<3, float> VectorR3f;
typedef VectorR<4, float> VectorR4f;

///////////////////////////////////////////////////////////////////////////////
// convert Euler angles(x,y,z) to axes(left, up, forward)
// Each column of the rotation matrix represents left, up and forward axis.
// The order of rotation is Roll->Yaw->Pitch (Rx*Ry*Rz)
// Rx: rotation about X-axis, pitch
// Ry: rotation about Y-axis, yaw(heading)
// Rz: rotation about Z-axis, roll
//    Rx           Ry          Rz
// |1  0   0| | Cy  0 Sy| |Cz -Sz 0|   | CyCz        -CySz         Sy  |
// |0 Cx -Sx|*|  0  1  0|*|Sz  Cz 0| = | SxSyCz+CxSz -SxSySz+CxCz -SxCy|
// |0 Sx  Cx| |-Sy  0 Cy| | 0   0 1|   |-CxSyCz+SxSz  CxSySz+SxCz  CxCy|
///////////////////////////////////////////////////////////////////////////////
static inline void anglesToAxes(
  const VectorR3f angles,
  VectorR3f& left,
  VectorR3f& up,
  VectorR3f& forward)
{
  float sx, sy, sz, cx, cy, cz, theta;

  // rotation angle about X-axis (pitch)
  theta = (float)(angles[X] * PDP_DEG2RAD);
  sx = sinf(theta);
  cx = cosf(theta);

  // rotation angle about Y-axis (yaw)
  theta = (float)(angles[Y] * PDP_DEG2RAD);
  sy = sinf(theta);
  cy = cosf(theta);

  // rotation angle about Z-axis (roll)
  theta = (float)(angles[Z] * PDP_DEG2RAD);
  sz = sinf(theta);
  cz = cosf(theta);

  // determine left axis
  left[X] = cy*cz;
  left[Y] = sx*sy*cz + cx*sz;
  left[Z] = -cx*sy*cz + sx*sz;

  // determine up axis
  up[X] = -cy*sz;
  up[Y] = -sx*sy*sz + cx*cz;
  up[Z] = cx*sy*sz + sx*cz;

  // determine forward axis
  forward[X] = sy;
  forward[Y] = -sx*cy;
  forward[Z] = cx*cy;
}

struct RegionFaceR3f
{
  VectorR3f p[4];
};

} //namespace pdp
