/** @file
 *
 * Math utility classes and functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <algorithm>
#include <cmath>
#include <functional>
#include <stdexcept>
#include <iterator>
#include <vector>
#include <numeric>
#include <utility>
#include <limits>
#include <string>
#include <cstring>
#include <type_traits>
#include <elements/export.h>

namespace pdp {

const double PDP_EPSILON = 1.0e-6;
const double PDP_PI      = 3.141592653589793;
const double PDP_HALFPI  = 1.570796326794897;
const double PDP_TWOPI   = 6.283185307179586;
const double PDP_DEG2RAD = PDP_TWOPI / 360.0;
const double PDP_RAD2DEG = 360.0 / PDP_TWOPI;

typedef unsigned int Dimension;
const Dimension X = 0;
const Dimension Y = 1;
const Dimension Z = 2;

inline bool almostZero(
  float a,
  float epsilon = PDP_EPSILON)
{ return(std::fabs(a) < epsilon); }

/**
 * Checks equality between two doubles.  Note that doubles have 16 decimal
 * points of precision but the Dicom Double String (DS) can only encode 8
 * decimal points.  Generally this would mean an epsilon < .00000001 may not
 * make sense.
 *
 * @ingroup elements
 */
inline bool almostEqual(
  float a,
  float b,
  float epsilon = PDP_EPSILON)
{ return(almostZero(a-b, epsilon)); }

/**
 * Fast floor is an optimization of the std::floor function recommended by the
 * shark profiler in MacOS.  Use only when the input is expected to be in the
 * range -4.5e15 to 4.5e15.
 *
 * @ingroup elements
 */
inline double fastfloor(
  double f)
{
  double twoTo52 = 4503599627370496.0;
  double c = (f > 0.0 ? -twoTo52 : twoTo52);
  double result = (f-c)+c;
  if (f < result) result -= 1.0;
  return result;
}

inline int floorInt(
  double x)
{
  return static_cast<int>(fastfloor(x));
}

inline int nearestInt(
  double x)
{
  return static_cast<int>(fastfloor(x + 0.5));
}

inline double nearestIntd(
  double x)
{
  return floor(x + 0.5);
}

inline double floord(
  double x)
{
  return floor(x);
}

inline double ceild(
  double x)
{
  return ceil(x);
}

inline unsigned int nextPowerOfTwo(
  unsigned int n)
{
  unsigned int result = 1;
  while (result < n)
  {
    result <<= 1;
  }
  return result;
}

/**
 * Returns the maximum value possible in a specific data type
 */
template<typename E>
inline E containerMaxValue()
{
  return std::numeric_limits<E>::max();
}

/**
 * Returns the minimum value possible in a specific data type. Note use this in
 * place of std::numeric_limits::lower() when the most negative double/float value
 * possible is desired.
 */
template<typename E>
inline E containerMinValue()
{
  return std::numeric_limits<E>::lower();
}

template<>
inline double containerMinValue()
{
  return -std::numeric_limits<double>::max();
}

template<>
inline float containerMinValue()
{
  return -std::numeric_limits<float>::max();
}

template< typename E>
inline std::pair<E, E> pdpMinMax(
  const E* begin,
  const E* end)
{
  typedef E PixelType;
  PixelType minValue = containerMaxValue<PixelType>();
  PixelType maxValue = containerMinValue<PixelType>();
  for (const PixelType* itr = begin; itr != end; ++itr)
  {
    minValue = (*itr < minValue) ? *itr : minValue;
    maxValue = (*itr > maxValue) ? *itr : maxValue;
  }
  return std::make_pair(minValue, maxValue);
}

/**
 * Compares float and double numbers to be within their machine epsilon.
 * Machine epsilon is the smallest representable difference between two numbers
 * in float/double.
 *
 * Template prototype for compare exact.  Only handles arithmetic types, and
 * thought is not intended for non real types, does handle them.
 *
 * If the two compare real values differ in type, the called must cast one
 * to the other.  The function cannot make a reasonable decision on which way
 * to cast.  Promoting both to double, will cause the function to likely return
 * failure, since float cannot handle the precision of double.  However casting
 * down may cause equality when the answer is actually wrong.
 */
template <typename E, typename F>
inline bool exactEqual(
  E a,
  F b)
{
  // We handle integer types here too though it isn't the function's intent
  static_assert(std::is_same<E, F>::value == true, "Both types must be float or double, cast one argument to fix");
  static_assert(std::is_arithmetic<E>::value == true, "This function is only defined for float/double");

  if (std::numeric_limits<E>::is_exact)  // Is type handles exact equality (int/short etc)
  {
#ifndef _WINDOWS
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wfloat-equal"
#endif
    return (a == b);
#ifndef _WINDOWS
#pragma GCC diagnostic pop
#endif
  }
  else  // If type is in exact, we compare against machine epsilon for type
  {
    auto diff = a - b;
    diff = (diff < 0) ? -diff : diff;  //absolute value
    return (diff < std::numeric_limits<E>::epsilon());
  }
}

/**
 * Returns a string with enough significant figures to fully represent a unique
 * float value.  Converting to this string, and back should with no loss of
 * data
 * @ingroup elements
 */
PDPELEMENTS_EXPORT std::string floatToString(
  const float value);

  /**
   * Returns a string with enough significant figures to fully represent a unique
   * double value.  Converting to this string, and back should with no loss of
   * data
   * @ingroup elements
   */
PDPELEMENTS_EXPORT std::string doubleToString(
  const double value);

PDPELEMENTS_EXPORT void averageBufferBy2(
  int width,
  int height,
  unsigned char*& buffer);

/**
 * This function returns the median value or values.  One value is returned if
 * the sequence is odd, and 2 are returned if the sequence is even.  The caller
 * should identify sequences with an even number of sequence and decide what
 * to do with the two median values.
 *
 * This algorithm uses std::nth_element to find the median so this runs in
 * linear time.
 *
 * @throws logic_error if begin == end, or size of container is 0
 *
 * @ingroup elements
 */
template <typename RandAccessIter>
std::pair< typename std::iterator_traits<RandAccessIter>::value_type,
  typename std::iterator_traits<RandAccessIter>::value_type > medianValues(
    RandAccessIter begin,
    RandAccessIter end)
{
  auto size = std::distance(begin, end);
  if ((begin == end) || (size == 0))
    throw std::logic_error("medianValues: Requires a valid sequence with non zero size.");

  // Create a working copy of the values for nth_element
  std::vector< typename std::iterator_traits<RandAccessIter>::value_type > tmpVector(begin, end);
  auto tbegin = tmpVector.begin();
  auto tend = tmpVector.end();

  auto middle = size/2;  //integer division
  auto target = tbegin + middle;

  std::nth_element(tbegin, target, tend);

  if((size % 2) == 0)   // Even - return middle two
  {
    auto targetNeighbor = std::max_element(tbegin, target);
    return std::make_pair( *targetNeighbor, *target );
  }
  else                  // Odd - return middle
  {
    return std::make_pair( *target, *target );
  }
}

/**
 * Performs the same algorithm as medianValues but will average the median
 * values in a sequence with an even number of elements.  For an odd number
 * of elements, both functions provide the same result.
 * @note Sending this function an empty container will cause an exception
 *
 * @throws logic_error if length is 0 or begin == end
 *
 * @ingroup elements
 */
template <class RandAccessIter>
double medianValue(
  RandAccessIter begin,
  RandAccessIter end)
{
  size_t length = distance(begin, end);
  if (length == 0)
  {
    throw std::logic_error("medianValue: Empty list, no median");
  }

  auto median  = medianValues(begin, end);

  // check for non - zero length first
  if ((length % 2) == 0)
  {
    // if even, average the median values
    return (median.first + median.second) / 2.0;
  }
  else
  {
    // if odd, return the median value
    return median.first;
  }
}

} // namespace pdp

