/** @file
 * 
 * A collection of three dimensional surface represented by connected triangles, with material properties
 * assigned to each survace used to represent solid models for rendering
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <vector>
#include <elements/VectorR.h>
#include <elements/TriangleListR.h>
#include <models/export.h>

namespace pdp {

/**
 * A collection of three dimensional surface represented by connected triangles, with material properties
 * assigned to each survace used to represent solid models for rendering
 *
 * @ingroup models
 */
class PDPMODELS_EXPORT ModelR3f
{
public:
  /**
   * Model part structure
   */
  struct ModelPart
  {
    std::string name;
    Material material;
    TriangleListR3f surface;
  };

  struct ModelFace
  {
    VectorR3i verticiesIdx;
    VectorR3i normalsIdx;
  };

  struct ModelPartLookup
  {
    std::string name;
    Material material;
    std::vector<ModelFace> faces;
  };

  /**
   * ModelR3f constructor
   */
  ModelR3f();

  /**
   * ModelR3f destructor
   */
  ~ModelR3f();

  /**
   * Converts in the input OBJ file (text format) into a binary format for read and storage efficiency
   */
  bool convertObjToMrf(
    const char* fileprefix);

  /**
   * Converts in the input OBJ file (text format) into a binary format and saves to the input ostream
   */
  bool convertObjToMrf(
    FILE* inMtlFileptr,
    FILE* inObjFileptr,
    std::iostream& outputStream);

  /**
   * Reads a the model from the specified input OBJ file (text format)
   */
  bool loadFromObj(
    const char* fileprefix);

  /**
   * Reads a binary version of the model from the specified input file
   */
  void readBinaryFile(
    const char* filename);

  /**
   * Writes a binary version of the model to the specified output file
   */
  void writeBinaryFile(
    const char* filename);

  /**
   * Reads a binary version of the model from the specified binary mrf input file
   */
  void loadFromMrf(
    const char* fileprefix);

  /**
   * Reads a binary version of the model from the specified input stream
   */
  void loadFromMrf(
    std::iostream& inputStream);

  /**
   * Returns a contant reference to the parts list of the model
   */
  const std::vector<ModelPart*>& parts() const
  { return m_parts; }

  /**
   * Returns the part index of the part in the model with the input name or -1 if not match is found
   */
  void part(
    const std::string& name,
    std::vector<int>& partIdxs);

  /**
   * Returns all the part indexes
   */
  void parts(
    std::vector<int>& partIdxs);

  //---------------------------------------------------------------------------
  // Friend functions
  //--------------------------------------------------------------------------- 
  /**
   * The method gets the next ModelR3f from the input stream and
   * decodes it from little endian format.
   */
  friend void getModelR3f(
    std::istream& inputStream,
    ModelR3f& m)
  {
    // Get the parts
    unsigned int numParts = getR<unsigned int>(inputStream);
    m.m_parts.clear();
    ModelPart* part = new ModelPart();
    for (unsigned int i = 0; i < numParts; i++)
    {
      getStdString(inputStream, part->name);
      getMaterial(inputStream, part->material);
      getTriangleListR(inputStream, part->surface);
      m.m_parts.push_back(part);
    }
  }

  /**
   * The method encodes the input ModelR3f into little endian format
   * and writes it to the output stream.
   */
  friend void putModelR3f(
    const ModelR3f& m,
    std::ostream& outputStream)
  {
    // Write the parts
    putR<unsigned int>(static_cast<unsigned int>(m.m_parts.size()), outputStream);
    for (unsigned int i = 0; i < m.m_parts.size(); i++)
    {
      putStdString(m.m_parts[i]->name, outputStream);
      putMaterial(m.m_parts[i]->material, outputStream);
      putTriangleListR(m.m_parts[i]->surface, outputStream);
    }
  }

private:
  struct MaterialSpec
  {
    std::string name;
    Material material;
  };

private:
  void clear();

  void lookupMaterial(
    const std::string& name,
    const std::vector<MaterialSpec>& materials,
    ModelPartLookup& part);

private:
  std::vector<ModelPart*> m_parts;
};

} // namespace pdp

