/** @file
 *
 * Robot Interface
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2019
 */
#pragma once

#include <boost/asio.hpp>
#include <utilities/StoppableTask.h>
#include <models/export.h>

using boost::asio::ip::udp;

namespace pdp {

class Robot;

class PDPMODELS_EXPORT RobotPollingTask: public StoppableTask
{
public:
  RobotPollingTask(
    Robot& robot,
    unsigned int pollingIntervalInMilliSeconds)
    : m_robot(robot),
      m_pollingIntervalInMilliSeconds(pollingIntervalInMilliSeconds)
  {}

  virtual ~RobotPollingTask() {}

  void run();
  
private:
  Robot& m_robot;
  unsigned int m_pollingIntervalInMilliSeconds;
};

//=====================================================================================
// Robot manager Interface
//=====================================================================================
class PDPMODELS_EXPORT Robot
{
public:
  explicit Robot(
    std::function<void()>* commandResponseFunction,
    std::function<void(std::string)>* stateUpdateFunction,
    std::function<void(const unsigned char*, int, int)>* frameDisplayFunction,
    std::function<void()>* commandsTimeoutFunction,
    std::function<void()>* stateTimeoutFunction,
    std::function<void()>* videoTimeoutFunction)
    : m_commandResponseFunction(commandResponseFunction),
      m_stateUpdateFunction(stateUpdateFunction),
      m_frameDisplayFunction(frameDisplayFunction),
      m_commandsTimeoutFunction(commandsTimeoutFunction),
      m_stateTimeoutFunction(stateTimeoutFunction),
      m_videoTimeoutFunction(videoTimeoutFunction),
      m_pollingTask(nullptr),
      m_pollingThread(nullptr)
  {}

  bool connected() const
  { return (m_pollingTask != nullptr && m_pollingThread != nullptr); }

  virtual ~Robot() {};

  virtual bool connect(
    unsigned int pollingIntervalInMilliSeconds);

  virtual std::string log() = 0;

  virtual std::string parameters() = 0;

  virtual void setSlow() = 0;

  virtual void launch() = 0;

  virtual void land() = 0;

  virtual void stop() = 0;

  virtual void hover() = 0;

  /**
   * Manual control interface for robot navigation.
   * Control definitions:
   *   leftRight: Move left / right - range (-1.0f to 1.0f)
   *   backwardForward: Move backward / forward - range (-1.0f to 1.0f)
   *   downUp: Move down / up - range (-1.0f to 1.0f)
   *   ccwYawCcYaw: CCW yaw / CW yaw - range (-1.0f to 1.0f)
   */
  virtual void setVelocities(
    float leftRight,
    float backwardForward,
    float downUp,
    float ccwYawCcYaw) = 0;

  virtual void disconnect();

protected:
  friend class RobotPollingTask;

  virtual void ping() = 0;

protected:
  std::function<void()>* m_commandResponseFunction;
  std::function<void(std::string)>* m_stateUpdateFunction;
  std::function<void(const unsigned char*, int, int)>* m_frameDisplayFunction;
  std::function<void()>* m_commandsTimeoutFunction;
  std::function<void()>* m_stateTimeoutFunction;
  std::function<void()>* m_videoTimeoutFunction;

private:
  RobotPollingTask* m_pollingTask;
  std::thread* m_pollingThread;
};

} // namespace pdp
