/** @file
 *
 * Rendering engine for Drone flight controls.
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <boost/signals2.hpp>
#include <utilities/timeUtils.h>
#include <render/RenderContext.h>
#include <render/openglUtils.h>
#include <render/export.h>

using namespace std;
using namespace std::chrono;

namespace pdp {

class View;
class ViewLayout;
class ModelR3f;

/**
 * Rendering engine for Drone flight controls.
 * 
 * @ingroup render
 */
class PDPRENDER_EXPORT RayTracingEngine
{
public:
  /**
   * Constructs a RayTracingEngine
   */
  RayTracingEngine(
    int width,
    int height,
    bool createJpegImage);

  /**
   * Destructs a RayTracingEngine.
   */
  ~RayTracingEngine();

  void setEndpoint(
    const std::string& remoteEndpointAddress,
    unsigned short remoteEndpointPort)
  {
    m_remoteEndpointAddress = remoteEndpointAddress;
    m_remoteEndpointPort = remoteEndpointPort;
  }

  /**
   * Returns the current remote endpoint address
   */
  const std::string& remoteEndpointAddress() const
  { return m_remoteEndpointAddress; }

  /**
   * Returns the current remote endpoint port
   */
  unsigned short remoteEndpointPort() const
  { return m_remoteEndpointPort; }

  /**
   * Returns the current engine information
   */
  std::string info() const
  { return m_info; }

  RenderContext* context()
  { return m_context; }

  /**
   * Returns the current context information
   */
  std::string contextInfo() const
  { return m_context->glInfo(); }

  /**
   * Returns 'true' if rendering is enabled meaning that the context is valid
   */
  bool renderingEnabled() const
  { return m_context->renderingEnabled(); }

  /**
   * Renders the current state of the flight controls
   */
  void render(
    ModelR3f* model,
    float roll,
    float pitch,
    float yaw);

  /**
   * Returns the current uncompressed image buffer.
   */
  const unsigned char* uncompressedImage()
  { return m_uncompressedImage; }

  int width() const
  { return m_context->width(); }

  int height() const
  { return m_context->height(); }

  const unsigned char* jpegImage() const
  { return m_jpegImage; }

  unsigned int jpegImageLength() const
  { return m_jpegImageLength; }

  float renderTime() const
  { return (float)m_renderTime.count(); }

  float compressionTime() const
  { return (float)m_compressionTime.count(); }

private:
  std::string m_remoteEndpointAddress;
  unsigned short m_remoteEndpointPort;
  RenderContext* m_context;
  View* m_view;
  ViewLayout* m_viewLayout;
  std::string m_info;
  unsigned char* m_uncompressedImage;
  bool m_createJpegImage;
  unsigned char* m_jpegImage;
  unsigned int m_jpegImageLength;
#ifdef _WINDOWS
  chrono::duration<double, std::milli> m_renderTime;
  chrono::duration<double, std::milli> m_compressionTime;
  chrono::high_resolution_clock::time_point m_lastRenderTime;
#else
  std::chrono::duration<double, std::milli> m_renderTime;
  std::chrono::duration<double, std::milli> m_compressionTime;
  std::chrono::high_resolution_clock::time_point m_lastRenderTime;
#endif
};

} // namespace pdp

