/** @file
 *
 * View Class used to implement Opengl based views for rendering
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <imaging/VolumeR3.h>
#include <imaging/ImageR2.h>
#include <elements/RegionR.h>
#include <imaging/Image2dDetectionsAdaptor.h>
#include <elements/MatrixR4x4f.h>
#include <render/MouseEvent.h>
#include <render/export.h>

namespace pdp {

class RenderContext;
class ModelR3f;
class Agent2d;

/**
 * View Class used to implement Opengl based views for rendering
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT View
{
public:
  /**
   * Enumerated type used to identify the view type.
   */
  enum ProjectionType
  {
    ProjectionTypeUndefined = 0,
    ProjectionTypeOrthographic,
    ProjectionTypePerspective
  };

  /**
   * Constructs a new View.
   */
  View(
    const VectorR2i& position,
    const VectorR2i& size,
    RenderContext* context);

  /**
   * Destructs this View.
   */
  ~View()
  {}

  ProjectionType projectionType() const
  { return m_projectionType; }

  float zoom() const
  { return m_zoom; }

  /**
   * Sets the current zoom factor of the view. This method
   * is meant to be used by the ZoomPanController2D controller and will
   * not automatically trigger a display update for the view
   */
  void setZoom(
    float zoom)
  { m_zoom = zoom; }

  /**
   * Returns the current default position for pan / zoom resets
   */
  VectorR2i defaultPosition() const
  { return m_defaultPosition; }

  float defaultRotation() const
  { return m_defaultRotation; }

  /**
   * Returns the current position of the view in scene coordinates
   */
  VectorR2f position() const
  { return m_position; }

  float rotation() const
  { return m_rotation; }

  /**
   * Sets the position of the view in scene coordinates. This method updates
   * the view transform of the view to reposition the view within the
   * scene.
   */
  void setPosition(
    const VectorR2f& position)
  { m_position = position; }

  void setRotation(
    float rotation)
  { m_rotation = rotation; }

  const RegionR2i& viewport() const
  { return m_viewport; }

  RenderContext* context()
  { return m_context; }

  const RenderContext* context() const
  { return m_context; }

  const MatrixR4x4f& modelviewXform() const
  { return m_modelviewXform; }

  const MatrixR4x4f& modelviewInvXform() const
  { return m_modelviewInvXform; }

  const MatrixR4x4f& projectionXform() const
  { return m_projectionXform; }

  const MatrixR4x4f& projectionInvXform() const
  { return m_projectionInvXform; }

  const MatrixR4x4f& viewportXform() const
  { return m_viewportXform; }

  const MatrixR4x4f& viewportInvXform() const
  { return m_viewportInvXform; }

  /**
   * Sets the modelview transform.
   */
  void setModelview(
    VolumeR3::RenderOrientation renderOrientation);

  void setImage2dModelview();

  /**
   * Sets a perspective modelview transform.
   */
  void setPerspectiveModelview(
    float roll,
    float pitch,
    float yaw);

  /**
   * Reposition this 2D View based on the input tangent vector. The view is translated
   * from its current position by the input vector in a plane orthogonal to the
   * view direction. Two vectors, in viewport coordinates, are input which point
   * to the begin and end of the tangent vector.
   */
  void reposition(
    VectorR2f tangentStart,
    VectorR2f tangentEnd);

  void updateRotation(
    VectorR2f tangentStart,
    VectorR2f tangentEnd);

  /**
   * Sets the projection transform.
   */
  void setProjection(
    const VectorR3f& sizeSCE,
    VolumeR3::RenderOrientation renderOrientation,
    ProjectionType projectionType);

  /**
   * Sets the projection transform for 2D images.
   */
  void setImage2dProjection(
    const VectorR2f& sizeSCE);

  /**
   * Sets a perspective projection transform.
   */
  void setPerspectiveProjection();

  /**
   * Sets the viewport transform for use by the rendering agents.
   */
  void setViewport(
    const VectorR2i& position,
    const VectorR2i& size);

  float aspectRatio() const
  { return m_aspectRatio; }

  void addAgent(
    Agent2d* agent)
  { m_agents.push_back(agent); }

  /**
   * Render the slice specified by the input orientation and position. The input
   * parameter 'positionRatio' is a value between 0 and 1 which navigates from the near
   * plan (0.0) to the far plane (1.0) of the specified slice orientation. Values less then
   * 0.0 are set to 0.0 and values greater than 1.0 are set to 1.0
   */
  void renderSlice(
    VolumeR3::VolumeChannel channel,
    const VectorR2f& contextOffset,
    VolumeR3::RenderOrientation renderOrientation,
    const VectorR3f& displayPointSCE,
    VolumeR3::StructureType structureType,
    float structureAlpha);

  void renderImage(
    ImageR2::ImageChannel channel,
    const VectorR2f& contextOffset,
    const VectorR2f& displaySample);

  void renderDetectionsImage(
    const VectorR2f& contextOffset,
    const VectorR2f& displaySample);

  void renderDetections(
    Image2dDetectionsAdaptor* imageDetections,
    ImageR2::StructureChannel structureChannel,
    const std::vector<pdp::Color>& classColors,
    const std::vector<bool>& showClasses,
    const VectorR2f& displaySample);

  /**
   * Render the MIP on the currently loaded Volume Texture.
   */
  void renderMip(
    VolumeR3::VolumeChannel channel,
    const VectorR2f& contextOffset,
    VolumeR3::StructureType structureType,
    float structureAlpha,
    float density,
    float brightness);

  /**
   * Render the model.
   */
  void renderModel(
    ModelR3f* model,
    MatrixR4x4f& modelXfrom);

  /**
   * Handles mouse events occurring in the viewport of the view.
   * The coordinates are specified in the viewport coordinate system.
   * If this view accepts the event, it should return true. Otherwise,
   * the event will be sent on.
   */
  bool mouseEvent(
    const MouseEvent& event,
    std::function<void()> updateRender);

private:
  void translate(
    const VectorR2f& v)
  { add(m_position, v, m_position); }

private:
  RenderContext* m_context;
  ProjectionType m_projectionType;
  RegionR2i m_viewport;
  float m_aspectRatio;
  float m_zoom;
  VectorR2i m_defaultPosition;
  VectorR2f m_position;
  float m_defaultRotation;
  float m_rotation;
  MatrixR4x4f m_modelviewXform;
  MatrixR4x4f m_modelviewInvXform;
  MatrixR4x4f m_projectionXform;
  MatrixR4x4f m_projectionInvXform;
  MatrixR4x4f m_viewportXform;
  MatrixR4x4f m_viewportInvXform;
  VectorR3f m_xHat;
  VectorR3f m_yHat;
  VectorR3f m_zHat;
  std::vector<Agent2d*> m_agents;
};

} // namespace pdp

