/** @file
 *
 * OpenGL Utility functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#if defined(PDP_OS_UBUNTU)
#include <EGL/egl.h>
#include <EGL/eglext.h>
#include <GLES3/gl32.h>
#else // MacOSX and Windows
#include <GLFW/glfw3.h>
#include <render/opengl460Loader.h>
#endif
#ifdef PDP_NVIDIA_LOCAL
#include <cuda_gl_interop.h>
#endif
#include <elements/Color.h>
#include <elements/pdpmath.h>
#include <render/export.h>

namespace pdp {

class Color;

/**
 * Write the OpenGL information to the input string @a info. Results of this
 * operation are reported to the @a glInfo string
 * @ingroup render
 */
PDPRENDER_EXPORT void reportOpenGLInfo(
  std::string& glInfo);

bool PDPRENDER_EXPORT checkForGLExtension(
  const std::string& extension);

/**
 * Initialize the available OpenGL shading language. Results of this
 * operation are reported to the @a glInfo string
 * @ingroup render
 */
bool PDPRENDER_EXPORT initShadingLanguage(
  std::string& glInfo,
  std::pair<int, int> highestGLVersion);

/**
 * Check the OpenGL error condition. This function returns true if an error
 * conditions exists or false if there are no current errors. All GL errors
 * are recorded to the internal error string and available vai a call to the
 * @a getGLError function. The OpenGL error state returns to @a GL_NO_ERROR
 * when the call returns.
 * @ingroup render
 */
bool PDPRENDER_EXPORT checkGLError();

/**
 * Clears the OpenGL error que without reporting.
 * @ingroup render
 */
void PDPRENDER_EXPORT clearGLError();

/**
 * This function copies the current GL errors to the @a glInfo string. The
 * internally recorded error string is cleared when the call returns.
 * @ingroup render
 */
void PDPRENDER_EXPORT getGLErrors(
  std::string& glInfo);

/**
 * This function copies the current GL messages to the @a glInfo string. The
 * internally recorded message string is cleared when the call returns.
 * @ingroup render
 */
std::string PDPRENDER_EXPORT getGLMessages();

/**
 * This function appends the input string @a glInfo to the internally recorded
 * message string. The caller can optionally choose to append the current OpenGL
 * error string to the message. If this is done the current error string is cleared
 * when the function returns.
 * @ingroup render
 */
void PDPRENDER_EXPORT appendGLMessages(
  const std::string& glInfo,
  bool appendGLErrors = false);

/**
 * This function clears the internally recorded message string.
 * @ingroup render
 */
void PDPRENDER_EXPORT clearGLMessages();

} // namespace pdp

