#pragma once

#include <memory>
#include <map>
#include <stdexcept>
#include <boost/noncopyable.hpp>
#include <boost/algorithm/string.hpp>
#include <string>
#include <iostream>
#include <ostream>
#include <cstddef>
#include <utilities/export.h>

namespace pdp {
  
enum struct PDPUTILITIES_EXPORT LogLevel
{
  LogInfo,
  LogWarning,
  LogError
};

enum struct PDPUTILITIES_EXPORT LoggerType
{
  LogDatabase = 0,  // Write to database log file
  LogDevConsole     // Write to Visual studio or XCode console
};

/**
 * Provides a interface to access the application logger system.
 * The concrete implementation will be a thread safe logger that may be called
 * from any thread in the application.
 * @ingroup utilities
 */
class PDPUTILITIES_EXPORT Logger
{
public:
  Logger() = default;

  virtual ~Logger() = default;

  /**
   * Append a message to the log. The string is expected in utf8 encoding.
   */
  virtual void logMessage(
    const std::string& message, 
    LogLevel level = LogLevel::LogInfo) = 0;

  /**
   * Signifies a null object.  If true, the log function will become a no-op
   */
  virtual bool isNull() = 0;

  Logger(const Logger&) = delete;

  Logger& operator=(const Logger&) = delete;
};

/**
 * Initialize the logger system with the directory to log to.  If not called
 * first a null object will be returned for the logger
 */
PDPUTILITIES_EXPORT void initLogging(
  const std::string& directory);

/**
 * Returns the correct static instance of the logger.  Note if init Logging
 * is not called, this function will return a nullptr.
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT Logger* getLogger(
  LoggerType channel);

/**
 * Write a record to database.log of severity "error"
 */
PDPUTILITIES_EXPORT void logDatabaseError(
  const std::string& message,
  pdp::LogLevel level = pdp::LogLevel::LogError);

/**
 * Write a record to development console of severity "error"
 * @param blocking Causes a sleep for 200 ms to flush to the console
 */
PDPUTILITIES_EXPORT void logConsoleError(
  const std::string& message,
  bool blocking = false);
  
PDPUTILITIES_EXPORT void cleanupLoggers();

} // namespace pdp

